#!/usr/bin/env amspython
from scm.plams import *
import numpy as np
import os
import matplotlib.pyplot as plt
from typing import List
import random

"""
This script performs eReaxFF MD on graphite-type materials.  It is based on https://doi.org/10.1149/1945-7111/aca362

This example is only compatible with AMS2023 or later.

Modify ffield below to give the correct absolute path to the force field file.
"""

# The force field for eReaxFF
ffield = os.path.abspath('ffield_eReaxFF_modified') # just modified alpha parameter to match AMS internal equations 
symbol_e = 'El' # The symbol for an electron
initial_structure = 'structures/gr_2Li.xyz' # The initial graphite-type materials
n_cations = 2 # Initial total charge, i.e. if Li+ are present 
# Here the script will start from a structure with total charge n_cations and further add electrons up to ne_max
ne_max = n_cations # Maximum number of electrons to dope
nprocs = 24

def main():
    assert check_version(), f"You need to run AMS2023 or later for this example"

    init()

    original_molecule = Molecule(initial_structure)
    for ne in range(ne_max+1):
        run_jobs('mol'+str(ne), ne, original_molecule)

    finish()

def check_version() -> bool:
    try:
        from scm.version import release
        assert release >= '2022.2'
    except (ImportError, AssertionError):
        return False
    return True


def add_electrons(mol:Molecule) -> Molecule:
    """
    Returns a new Molecule where
    (1) Replace all C atoms to B-type
    (2) Add one electron for each B atom (slightly shifted by 0.35 Ang)
    """

    # Replace C to B (need to use ReaxFF parameters of B == C+)
    mol = mol.copy()
    for atom in mol.atoms:
        if atom.symbol == 'C':
            atom.symbol = 'B'

    # Create electrons
    # Duplicate and shift by [0.35,0,0] Angstrom
    for atom in mol.atoms:
        if atom.symbol == 'B':
            coords_e = (atom.coords[0]+0.35,atom.coords[1],atom.coords[2])
            mol.add_atom(Atom(symbol=symbol_e, coords=coords_e))

    return mol

def add_extra_electrons(mol:Molecule, ne) -> Molecule:
    """
        Adds an electron next to the atom with index atom_index and decreases the charge
    """
    mol = mol.copy()
    nat = len(mol.atoms)

    if ne == 0: return mol

    try:
        ne <= nat 
    except ValueError:
        print('Change number of electrons.')

    for ei in range(ne):
        ref_coords = mol[ei+1].coords
        coords = ref_coords[0]-0.7, ref_coords[1], ref_coords[2]
        mol.add_atom(Atom(symbol=symbol_e, coords=coords))

#    try:
#        mol.properties.charge -= ne-n_cations
#    except TypeError:
#        mol.properties.charge = -ne+n_cations
    total_charge = n_cations-ne
    mol.properties.charge = total_charge 

    return mol

def get_constraints_list(mol:Molecule) -> List[str]:
    """ Returns a list that can be assigned to sett.input.reaxff.charges.constraint """

    clist = []
    sym2charge = {'B': 1, symbol_e: -1, 'Lv': 1}

    for i, atom in enumerate(mol.atoms, 1):
        if atom.symbol in sym2charge:
            region = f'{atom.symbol}{i}'
            charge = sym2charge[atom.symbol]

            atom.properties.region = region
            clist.append(f'region={region} charge={charge}')
            
    return clist

def get_fixedregion(mol:Molecule) -> List[str]:
    """ Return a list of 10% of the B atoms to fix """

    clist = []
    nfix = int(np.ceil(0.01*len(mol.atoms)))
    blist = [i for i,at in enumerate(mol, 1) if at.symbol == 'B']
    fixid = random.sample(blist, nfix) 
    activeid_B = [i for i in blist if i not in fixid]
    activeid_symbols = {}
    for i in activeid_B: activeid_symbols[i] = 'B'

    blist = [i for i,at in enumerate(mol, 1) if at.symbol == 'Lv']
    activeid_Lv = [i for i in blist]
    for i in activeid_Lv: activeid_symbols[i] = 'Lv'
    
    blist = [i for i,at in enumerate(mol, 1) if at.symbol == 'El']
    activeid_El = [i for i in blist]
    for i in activeid_El: activeid_symbols[i] = 'El'

    activeid = activeid_B + activeid_Lv + activeid_El

    s = Settings()
    s.input.ams.constraints.AtomList = ' '.join([str(i) for i in fixid]) 

    return s,activeid,activeid_symbols

def set_md(activeid,activeid_symbols) -> Settings:
    s = Settings()
    s.input.ams.task = 'MolecularDynamics'
    s.input.ams.MolecularDynamics.InitialVelocities.Temperature = 1.0
    s.input.ams.MolecularDynamics.Thermostat.Temperature = '1 300 300'
    s.input.ams.MolecularDynamics.Thermostat.Duration = '10000 490000'
    s.input.ams.MolecularDynamics.Thermostat.Type = 'NHC' 
    s.input.ams.MolecularDynamics.Thermostat.Tau = 100
    rlist = []
    for i in activeid: rlist.append(activeid_symbols[i]+str(i))
    s.input.ams.MolecularDynamics.Thermostat.Region = '+'.join(rlist) 
    s.input.ams.MolecularDynamics.Nsteps = 500000
    s.input.ams.MolecularDynamics.Timestep = 0.1
    s.input.ams.MolecularDynamics.Checkpoint.Frequency = 1000000
#    s.input.ams.MolecularDynamics.Preserve.CenterOfMass = 'yes' 

    return s

def set_geoopt():

    s = Settings()
    s.input.ams.task = 'GeometryOptimization'
#    s.input.ams.GeometryOptimization.MaxIterations = 100
#    s.input.ams.GeometryOptimization.OptimizeLattice = 'Yes'
#    s.input.ams.GeometryOptimization.PretendConverged = 'Yes'
    return s

def general_settings() -> Settings:
    s = Settings()
    s.runscript.nproc = nprocs
#    s.runscript.preamble_lines = [f'export OMP_NUM_THREADS=1']
    s.input.reaxff.forcefield = ffield
    return s

def run_jobs(name:str, ne:int, original_molecule:Molecule) -> float:
    """ Returns the electron affinity in kcal/mol """
    # Neutral
    # Add electrons, charges and constrains
    s = general_settings()
    mol = add_electrons(original_molecule)
    mol = add_extra_electrons(mol, ne)
    total_charge = n_cations-ne
    s.input.ams.system.charge = total_charge 
    s.input.reaxff.charges.constraint = get_constraints_list(mol)
    sfix,activeid,activeid_symbols = get_fixedregion(mol) 
    s += sfix 

    # Optimization geometry
    j = AMSJob(molecule=mol, name=f'{name}_preopt', settings=s+set_geoopt())
    res = j.run(watch=True)

    mol_preopt = j.results.get_main_molecule()

    # Run MD 
    j = AMSJob(molecule=mol_preopt, name=f'{name}_md', settings=s+set_md(activeid,activeid_symbols))
    res = j.run(watch=True)
    
    mol_md = j.results.get_main_molecule()
    
    # Optimize geometry
    j = AMSJob(molecule=mol_md, name=f'{name}_opt', settings=s+set_geoopt())
    res = j.run(watch=True)

if __name__ == '__main__':
    main()

